﻿using Cosmos.HAL.Audio;
using Cosmos.HAL.Drivers.PCI.Audio;
using Cosmos.System.Audio;
using Cosmos.System.Audio.DSP.Processing;
using Cosmos.System.Audio.IO;
using IL2CPU.API.Attribs;
using System;
using Sys = Cosmos.System;

namespace CAIDemo
{
    /// <summary>
    /// A demo kernel that demonstrates the usage of the Cosmos Audio Infrastructure.
    /// </summary>
    public class Kernel : Sys.Kernel
    {
        [ManifestResourceStream(ResourceName = "CAIDemo.SampleAudio.wav")]
        private readonly static byte[] sampleAudioBytes;

        // An audio manager allows for interfacing between different
        // audio drivers with a single object.
        AudioManager audioManager;

        // An audio driver handles communication between the kernel and a
        // sound card.
        AudioDriver driver;

        // An audio mixer is an audio stream that serves a purpose of
        // mixing several audio streams together.
        AudioMixer mixer;

        // A seekable audio stream is a traditional stream that has a length
        // and a position property. An example of an audio stream that is not
        // seekable would be, for example, an audio mixer stream, or a microphone
        // input stream.
        SeekableAudioStream audioStream;

        protected override void BeforeRun()
        {
            Console.WriteLine("[kernel] Initializing audio driver...");

            try
            {
                // The Initialize method creates a new global instance of the
                // AC97 driver if it does not exist yet. This is to ensure
                // that only one instance of the AC97 object is ever created
                // throught the execution of the kernel.
                driver = AC97.Initialize(4096);
            }
            catch (InvalidOperationException)
            {
                // All audio drivers will throw an InvalidOperationException
                // if there is an attempt to create a driver for a device
                // that is not present on the system.
                Console.ForegroundColor = ConsoleColor.Red;
                Console.WriteLine("[kernel] No AC97 device found.");

                while (true)
                {
                    Console.ReadKey(true);
                }
            }

            mixer = new AudioMixer();

            // The MemoryAudioStream has a built-in .WAV file parser -
            // however, creating a memory audio stream from raw audio
            // samples is also possible:
            //      audioStream = new MemoryAudioStream(
            //          new SampleFormat(AudioBitDepth.Bits16, 2, true), 48000, sampleAudioBytes
            //      );
            audioStream = MemoryAudioStream.FromWave(sampleAudioBytes);

            // The CAI also allows for post-processing effects. A simple and useful
            // gain adjustment post-processing effect is already available
            // out-of-the-box.
            audioStream.PostProcessors.Add(new GainPostProcessor(0.5f));

            // With an audio mixer, we can add and remove streams at any time.
            mixer.Streams.Add(audioStream);

            Console.WriteLine("[kernel] Driver OK, initializing audio manager...");

            // An audio manager will automatically handle communicating with the
            // audio driver. It's possible to directly communicate with the
            // driver without the use of an audio manager in order to achieve
            // more control over your kernel.
            audioManager = new AudioManager()
            {
                Stream = mixer,
                Output = driver
            };

            Console.WriteLine("[kernel] Initialization OK, enabling...");
            audioManager.Enable(); // Enabling the audio manager will automatically enable the driver.

            Console.ForegroundColor = ConsoleColor.DarkGray;
            Console.WriteLine("[kernel] The audio should be playing - press any key to see the position of the audio in bytes.");
        }

        protected override void Run()
        {
            // Playback of audio is completely asynchronous - you can interact
            // with your kernel while an audio stream is being played.

            // Please keep in mind that audio mixing may slow down the kernel,
            // especially when the played samples are of different formats.
            // While not recommended, the audio mixer accepts streams of
            // varying formats. If at least one audio stream has a different
            // format than the others, the mixer will first normalize the
            // streams to a common bit-depth, which will slow down the kernel.

            // It's recommended to perform bit-conversion operations offline
            // before passing them to an AudioMixer.

            Console.ReadKey(true);
            Console.WriteLine($"(position: {audioStream.Position} / {audioStream.Length})");
        }
    }
}
